import locale


def format_quantity(quantity):
    """
        format the quantity
    """
    if quantity is not None:
        result = locale.format_string('%g', quantity, grouping=True)
        if isinstance(result, bytes):
            result = result.decode('utf-8')
        return result
    else:
        return ""



"""
    Usefull functions
"""
import time
import datetime

from sqlalchemy import (
    func,
    cast,
)
from sqlalchemy.types import NullType
from sqlalchemy import BigInteger

DEFAULT_DATE = datetime.date(2000, 1, 1)


def format_to_taskdate(value):
    """
        format a datetime.date object to a 'taskdate' format:
        an integer composed from a string YYYYmmdd
        Sorry .... it's not my responsability
    """
    if value is None:
        return None
    elif isinstance(value, datetime.date):
        if value.year < 1900:
            value.year = 1901
        return int(value.strftime("%Y%m%d"))
    else:
        return int(value)


def get_current_timestamp():
    """
        returns current time
    """
    return int(time.time())


def non_null_sum(column, *args, **kwargs):
    """
    Cast the sqlalchemy built with func.sum to the same type as the column
    being summed

    Usefull to avoid getting decimal when we expect an int
    :param obj column: The SqlAlchemy column (e.g : Payment.amount)
    :returns: A SqlAlchemy query option (passed to Dbsession.query)
    """
    query = func.sum(column, *args, **kwargs)
    # Prevent Decimal typing when agregating
    query = func.ifnull(query, 0)
    if not isinstance(column.type, NullType):
        query = cast(query, column.type.__class__)
    else:
        query = cast(query, BigInteger)
    return query



"""
    Database access base objects
    DBSESSION : database session factory
    DBBASE : base object for models
"""
from sqlalchemy.ext import declarative
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm import scoped_session
from sqlalchemy.ext.declarative import declared_attr
from sqlalchemy.schema import MetaData
from zope.sqlalchemy import register
from endi_base.utils.ascii import camel_case_to_name

DBSESSION = scoped_session(sessionmaker())
register(DBSESSION)


class ORMClass(object):
    """
        Base class for our models providing usefull query and get methods
    """
    @classmethod
    def query(cls, *args):
        """
            return a query
        """
        if not args:
            return DBSESSION().query(cls)
        else:
            query_args = []
            for arg in args:
                cls_attr = getattr(cls, arg, None)
                if cls_attr is not None:
                    query_args.append(cls_attr)
            return DBSESSION().query(*query_args)

    @classmethod
    def get(cls, id_):
        """
            Return a query
        """
        return DBSESSION().query(cls).get(id_)

    @declared_attr
    def __tablename__(cls):
        return camel_case_to_name(cls.__name__)


NAMING_CONVENTION = {
    "ix": 'ix_%(column_0_label)s',
    "uq": "uq_%(table_name)s_%(column_0_name)s",
    "ck": "ck_%(table_name)s_%(column_0_name)s",
    "fk": "fk_%(table_name)s_%(column_0_name)s",
    "pk": "pk_%(table_name)s"
}

METADATA = MetaData(naming_convention=NAMING_CONVENTION)


DBBASE = declarative.declarative_base(
    cls=ORMClass,
    metadata=METADATA,
)


def record_to_appstruct(self):
    """
        Transform a SQLAlchemy object into a deform compatible dict
        usefull to autofill an editform directly from db recovered datas
    """
    return dict([(k, self.__dict__[k])
                for k in sorted(self.__dict__) if '_sa_' != k[:4]
                and self.__dict__[k] is not None])


DBBASE.appstruct = record_to_appstruct


default_table_args = {'mysql_engine': 'InnoDB', "mysql_charset": 'utf8mb4'}



"""
    Provide common tools for string handling
"""
import re
import random
import string
from unidecode import unidecode


def force_ascii(value):
    """
        Return enforced ascii string
        éko=>ko
    """
    if isinstance(value, bytes):
        value = force_string(value)
    elif not isinstance(value, str):
        # Supports numbers, for instance
        value = str(value)
    value = unidecode(value)
    return value


def force_utf8(value):
    """
    return a utf-8 byte string

    :param str value:
    :rtype: bytes
    """
    if isinstance(value, str):
        value = value.encode('utf-8')
    return value


def force_string(value):
    """
    return an utf-8 str

    :param bytes value: The original value to convert
    :rtype: str
    """
    if isinstance(value, bytes):
        value = value.decode('utf-8') 
    return value


force_unicode = force_string


def camel_case_to_name(name):
    """
    Used to convert a classname to a lowercase name
    """
    def convert_func(m):
        return "_" + m.group(0).lower()
    return name[0].lower() + re.sub(r'([A-Z])', convert_func, name[1:])


def gen_random_string(size=15):
    """
    Generate random string

        size

            size of the resulting string
    """
    return ''.join(random.choice(string.ascii_lowercase) for _ in range(size))


def random_tag_id(size=15):
    """
    Return a random string supposed to be used as tag id
    """
    return gen_random_string(size)


def to_utf8(datas):
    """
    Force utf8 string entries in the given datas
    """
    res = datas
    if isinstance(datas, dict):
        res = {}
        for key, value in datas.items():
            key = to_utf8(key)
            value = to_utf8(value)
            res[key] = value

    elif isinstance(datas, (list, tuple)):
        res = []
        for data in datas:
            res.append(to_utf8(data))

    elif isinstance(datas, str):
        res = datas.encode('utf-8')

    return res


def isint(val):
    """
    Return True if the result looks like an int
    """
    res = False

    if isinstance(val, int):
        res = True

    elif isinstance(val, (str, bytes)) and val.isdigit():
        res = True

    return res


def force_filename(val):
    """
    Transform a string to a valid filename
    """
    result = force_ascii(val)
    result = result.replace(' ', '_')

    valid_chars = "-_.%s%s" % (string.ascii_letters, string.digits)
    result = "".join((char for char in result if char in valid_chars))

    return result



class UndeliveredMail(Exception):
    """
    A custom class for undelivered emails
    """
    pass


class MailAlreadySent(Exception):
    """
    A custom class for mail that were already sent
    """
    pass



"""
    Custom types and usefull functions
"""
import simplejson as json

from sqlalchemy.ext.mutable import Mutable
from sqlalchemy.types import (
    TypeDecorator,
    Text as Text_type,
)
from pyramid.authorization import (
    Allow,
    ALL_PERMISSIONS,
)


class JsonEncodedDict(TypeDecorator):
    """
    Stores a dict as a json string in the database
    """
    impl = Text_type

    def process_bind_param(self, value, dialect):
        """
            Process params when setting the value
        """
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        """
            Processing the value when getting the value
        """
        if value is not None:
            value = json.loads(value)
        return value


class JsonEncodedList(TypeDecorator):
    """
    Stores a list as a json string
    """
    impl = Text_type

    def process_bind_param(self, value, dialect):
        """
            Process params when setting the value
        """
        if value is not None:
            value = json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        """
            Processing the value when getting the value
        """
        if value is not None:
            value = json.loads(value)
        return value


class ACLType(JsonEncodedList):
    all_permissions_serialized = '__ALL_PERMISSIONS__'
    defaults = [
        (Allow, 'group:admin', ALL_PERMISSIONS),
        (Allow, "group:manager", ("manage", "add", "edit", "view")),
    ]

    def process_bind_param(self, value, dialect):
        if value is not None:
            value = [list(ace) for ace in value if ace not in self.defaults]
            for ace in value:
                if ace[2] == ALL_PERMISSIONS:
                    ace[2] = self.all_permissions_serialized
        return JsonEncodedList.process_bind_param(self, value, dialect)

    def process_result_value(self, value, dialect):
        acl = JsonEncodedList.process_result_value(self, value, dialect)
        if acl is not None:
            for ace in acl:
                if ace[2] == self.all_permissions_serialized:
                    ace[2] = ALL_PERMISSIONS
            return self.defaults + [tuple(ace) for ace in acl]


class MutableDict(Mutable, dict):
    """
        Allows sqlalchemy to check if a data has changed in our dict
        If not used, the dbsession will never detect modifications

        Note : only associating a value to one key will work (no subdict
            handling)
    """
    @classmethod
    def coerce(cls, key, value):
        "Convert plain dictionaries to MutableDict."

        if not isinstance(value, MutableDict):
            if isinstance(value, dict):
                return MutableDict(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def __setitem__(self, key, value):
        "Detect dictionary set events and emit change events."

        dict.__setitem__(self, key, value)
        self.changed()

    def __delitem__(self, key):
        "Detect dictionary del events and emit change events."

        dict.__delitem__(self, key)
        self.changed()


class MutableList(Mutable, list):
    """
        Allows sqlalchemy to check if a data has changed in our list
        If not used, the dbsession will never detect modifications
    """
    @classmethod
    def coerce(cls, key, value):
        """
        Convert list to mutablelist
        """
        if not isinstance(value, MutableList):
            if isinstance(value, list):
                return MutableList(value)

            # this call will raise ValueError
            return Mutable.coerce(key, value)
        else:
            return value

    def append(self, value):
        """
        Detect list append changes
        """
        list.append(self, value)
        self.changed()

    def extend(self, value):
        """
        Detect list append changes
        """
        list.extend(self, value)
        self.changed()

    def remove(self, value):
        """
        Detect list remove change
        """
        list.remove(self, value)
        self.changed()


# Here we always associate our MutableDict to our JsonEncodedDict column type
# If a column is of type JsonEncodedDict, its value will be casted as a
# mutabledict that will signify modifications on setitem and delitem
MutableDict.associate_with(JsonEncodedDict)
# The same for lists
MutableList.associate_with(JsonEncodedList)



"""
Various models mixins adding new features/columns to SQLAlchemy models
"""
from datetime import datetime

from sqlalchemy import (
    Column,
    Integer,
    ForeignKey,
    DateTime,
    String,
)
from sqlalchemy.ext.declarative import declared_attr
from .types import (
    ACLType,
    MutableList,
)


class TimeStampedMixin(object):
    @declared_attr
    def created_at(cls):
        return Column(
            DateTime(),
            info={
                'colanderalchemy': {
                    'exclude': True, 'title': "Créé(e) le",
                }
            },
            default=datetime.now,
        )

    @declared_attr
    def updated_at(cls):
        return Column(
            DateTime(),
            info={
                'colanderalchemy': {
                    'exclude': True, 'title': "Mis(e) à jour le",
                }
            },
            default=datetime.now,
            onupdate=datetime.now
        )


class OfficialNumberMixin:
    """
    Models that implements that mixin are then able to be numbered using
    numbering sequence mechanisms with guarantees on non-duplicate.

    Models are required to be child of endi.models.node.Node

    Features :

    - templated creation of official_number that may use one or more numeric
      sequences
    - guarantees on official_number (unicity and no gap in sequences)
    - instances with an official_number set should not be deleted

    Inheriting class must:

    - ensure that once official_number is set, instances are not deleted

    Inheriting class may:

    - redefine official_number and company_id as long as name and type are kept
      as-is

    - redefine validation_date_column if the validation is not stored into
      self.date
    """

    validation_date_column = 'date'

    @declared_attr
    def official_number(cls):
        return Column(
            String(255),
            default=None,
        )

    @declared_attr
    def company_id(cls):
        return Column(
            Integer,
            ForeignKey('company.id'),
            nullable=False,
        )

    @property
    def validation_date(self) -> datetime.date:
        return getattr(self, self.validation_date_column)

    @classmethod
    def get_validation_date_column(cls) -> Column:
        # could be turned into @classmethod + @property on Py3.9
        return getattr(cls, cls.validation_date_column)


class PersistentACLMixin(object):
    """Extend pyramid ACL mechanism to offer row-level ACL

    Subclasses may set the following attributes :

    - ``.__default_acl__`` as an ACL or callable returning an ACL, traditionaly
      set at class level.
    - ``.__acl__`` as an ACL or callable returning an ACL, this can be set on
      class or directly on instance.

    If both are defined, ``__acl__``, the more specific, is prefered.

    Use actual callables, not properties, in subclasses to prevent pyramid from
    silently ignoring any AttributeError your callable should trigger.
    """
    @declared_attr
    def _acl(cls):
        """
        Customizable _acl column
        """
        return Column(
            MutableList.as_mutable(ACLType),
            info={
                'colanderalchemy': {'exclude': True},
                'export': {'exclude': True}
            },
        )

    def _get_acl(self):
        # Any AttributeError raised at this function level would be masked.
        # See https://github.com/Pylons/pyramid/pull/2613/files
        if getattr(self, '_acl', None) is None:
            if getattr(self, "__default_acl__", None) is not None:
                return self.__default_acl__
            elif getattr(self, "parent", None) is not None:
                return self.parent.__acl__
            else:
                raise AttributeError('__acl__')
        return self._acl

    def _set_acl(self, value):
        self._acl = value

    def _del_acl(self):
        self._acl = None

    __acl__ = property(_get_acl, _set_acl, _del_acl)


class DuplicableMixin(object):
    """
    Factorize duplication logic on sqlalchemy models

    __duplicable_fields__ attr should contain the list of attributes to be
    copied from a model to its duplicate.
    """
    __duplicable_fields__ = None

    def duplicate(self, factory=None, **kwargs):
        """
        Duplicate all fields listed in __duplicable_fields__ plus optional
        ad-hoc values.

        :param class factory: The new instance class
        :param kwargs: allow to provide ad-hoc values for the attrs of the
          new instance. They will take precedence over copied values.
        :returns: new instance combining fields copied from self, and ad-hoc
          args. The instance is not saved.
        """

        if self.__duplicable_fields__ is None:
            raise NotImplementedError(
               "DuplicableMixin implementors should define "
               + "__duplicable_fields__"
            )

        if factory is None:
            factory = self.__class__

        duplicate_obj = factory()
        for i in self.__duplicable_fields__:
            setattr(duplicate_obj, i, getattr(self, i))
        for k, v in kwargs.items():
            setattr(duplicate_obj, k, v)
        return duplicate_obj



"""
Date manipulation utilities
"""
import datetime


def str_to_date(
    str_date,
    formats=("%d/%m/%Y", "%d-%m-%Y", "%d/%m/%y", "%Y-%m-%d", "%Y%m%d")
):
    """
    Transform a date string to a date object

    :param str str_date: The date string
    :param tuple formats: List of date format to try when parsing
    :return: A datetime object
    :rtype: datetime.date
    """
    res = None
    if str_date is not None:
        for format_ in formats:
            try:
                res = datetime.datetime.strptime(str_date, format_)
            except ValueError:
                pass
            else:
                break
    return res


def get_strftime_from_date(date_obj, template_str):
    """
    Return the result of date.strftime(template_str) handling exceptions
    """
    try:
        resp = date_obj.strftime(template_str)
    except ValueError:
        resp = ""
    return resp


def format_short_date(date):
    """
        return a short printable version of the date obj
    """
    if isinstance(date, datetime.date):
        resp = get_strftime_from_date(date, "%d/%m/%Y")
    elif not date:
        resp = ""
    else:
        date_obj = datetime.datetime.fromtimestamp(float(date))
        resp = get_strftime_from_date(date_obj, "%d/%m/%Y %H:%M")
    return resp


def format_datetime(datetime_object, timeonly=False):
    """
    format a datetime object
    """
    res = get_strftime_from_date(datetime_object, "%H:%M")
    if not timeonly:
        day = get_strftime_from_date(datetime_object, "%d/%m/%Y")
        res = "%s à %s" % (day, res)
    return res


def format_long_date(date):
    """
        return a long printable version of the date obj
    """
    if isinstance(date, datetime.date):
        str_date = get_strftime_from_date(date, "%d %B %Y")
        resp = "{0}".format(str_date.capitalize())
    elif not date:
        resp = ""
    else:
        date = datetime.datetime.fromtimestamp(float(date))
        str_date = get_strftime_from_date(date, "%d %B %Y")
        resp = "{0}".format(str_date.capitalize())
    return resp


def format_date(date, short=True):
    """
        return a pretty print version of the date object
    """
    if short:
        return format_short_date(date)
    else:
        return format_long_date(date)


def format_duration(duration, short=True):
    """
    return a pretty print version of a duration.

    :param (int,int) duration: hours,minutes tuple to convert.
    :param bool short: if True, hide minutes part when it equals zero.
    """
    hours, minutes = duration
    if minutes == 0 and short:
        return '{}h'.format(hours)
    else:
        return '{}h{:02d}'.format(hours, minutes)



import datetime
import colander

from decimal import Decimal
from sqlalchemy import (
    Boolean,
    Date,
    DateTime,
    Integer,
    Float,
    Numeric
)
from pyramid.renderers import JSON


def set_export_formatters():
    """
    Globally set export formatters in the sqla_inspect registry
    """
    from sqla_inspect.export import FORMATTERS_REGISTRY
    from endi_base.utils.date import (
        format_date,
        format_datetime,
    )
    from endi_base.utils.strings import (
        format_quantity,
    )
    from endi_base.utils.export import format_boolean

    FORMATTERS_REGISTRY.add_formatter(
        Date, format_date, 'py3o'
    )
    FORMATTERS_REGISTRY.add_formatter(
        DateTime, format_datetime, 'py3o'
    )
    FORMATTERS_REGISTRY.add_formatter(
        Date, format_date, 'csv'
    )
    FORMATTERS_REGISTRY.add_formatter(
        DateTime, format_date, 'csv'
    )
    FORMATTERS_REGISTRY.add_formatter(Boolean, format_boolean)
    FORMATTERS_REGISTRY.add_formatter(Float, format_quantity)
    FORMATTERS_REGISTRY.add_formatter(Integer, format_quantity)
    FORMATTERS_REGISTRY.add_formatter(Numeric, format_quantity)


def set_export_blacklist():
    """
    Globally set an export blacklist
    """
    from sqla_inspect.export import BLACKLISTED_KEYS

    BLACKLISTED_KEYS.extend([
        '_acl',
        'password',
        'parent_id',
        'parent',
        'type_',
        'children',
    ])


def set_xls_formats():
    """
    Globally set the xls formats by datatype
    """
    from sqla_inspect.excel import FORMAT_REGISTRY

    FORMAT_REGISTRY.add_item(Date, "dd/mm/yyyy")
    FORMAT_REGISTRY.add_item(DateTime, "dd/mm/yyyy hh:mm")


def configure_export():
    """
    Customize sqla_inspect tools
    """
    set_export_formatters()
    set_export_blacklist()
    set_xls_formats()


def set_json_renderer(config):
    """
    Customize json renderer to allow datetime rendering
    """
    json_renderer = JSON()

    def toisoformat(obj, request):
        return obj.isoformat()

    json_renderer.add_adapter(datetime.datetime, toisoformat)
    json_renderer.add_adapter(datetime.date, toisoformat)
    json_renderer.add_adapter(colander._null, lambda _, r: "null")

    def decimal_to_num(obj, request):
        return float(obj)

    json_renderer.add_adapter(Decimal, decimal_to_num)

    config.add_renderer('json', json_renderer)
    return config



import logging
import urllib.parse
from pyramid_mailer import get_mailer
from pyramid_mailer.message import (
    Message,
)

logger = logging.getLogger(__name__)

UNSUBSCRIBE_MSG = "<mailto:{0}?subject=Unsubscribe-{1}>"


UNSUBSCRIBE_LINK = """


Vous avez reçu ce mail car vous êtes utilisateurs de l'application enDI. \
Si vous avez reçu ce mail par erreur, nous vous prions de nous \
en excuser. Vous pouvez vous désincrire en écrivant à \
{0}?subject=Unsubscribe-{1}."""


def format_mail(mail):
    """
    Format the mail address to fit gmail's rfc interpretation
    """
    return "<{0}>".format(mail)


def format_link(settings, link):
    """
    Format a link to fit the sender's domain name if a bounce url has been
    configured
    """
    bounce_url = settings.get("mail.bounce_url")
    if bounce_url:
        link = urllib.parse.quote(link)
        url = "http://{0}/?url={1}".format(bounce_url, link)
    else:
        url = link
    return url


def get_sender(settings):
    """
    Return the mail sender's address
    """
    if 'mail.default_sender' in settings:
        mail = settings['mail.default_sender']
    else:
        mail = "Unknown"
    return format_mail(mail)


def _handle_optout(settings, mail_body):
    """
    Add additionnal datas for optout declaration
    Allows to fit a bit more the mailing conformity
    """
    headers = {}
    optout_addr = settings.get("mail.optout_address")
    instance_name = settings.get('endi.instance_name')
    if optout_addr and instance_name:
        headers['Precedence'] = 'bulk'
        headers['List-Unsubscribe'] = UNSUBSCRIBE_MSG.format(
                optout_addr,
                instance_name,
                )
        mail_body += UNSUBSCRIBE_LINK.format(optout_addr, instance_name)
    return headers, mail_body


def send_mail(request, recipients, body, subject, attachment=None):
    """
    Try to send an email with the given datas

    :param obj request: a pyramid request object
    :param list recipients: A list of recipients strings
    :param str body: The body of the email
    :param str subject: The subject of the email
    :param obj attachment: A pyramid_mailer.message.Attachment object

    """
    if not isinstance(recipients, (tuple, list)):
        recipients = [recipients]

    if len(recipients) == 0:
        return False
    logger.info("Sending an email to '{0}'".format(recipients))
    settings = request.registry.settings
    headers, mail_body = _handle_optout(settings, body)
    try:
        recipients = [format_mail(recipient) for recipient in recipients]
        sender = get_sender(settings)
        mailer = get_mailer(request)
        message = Message(
            subject=subject,
            sender=sender,
            recipients=recipients,
            body=mail_body,
            extra_headers=headers
        )
        if attachment:
            message.attach(attachment)
        mailer.send_immediately(message)
    except Exception:
        import traceback
        traceback.print_exc()
        logger.exception(" - An error has occured while sending the \
email(s)")
        return False
    return True



"""
    Math utilities used for computing
"""
import math
from decimal import (
    Decimal,
    ROUND_HALF_UP,
    ROUND_DOWN,
)

PRECISION_LEVEL = 2


def floor_to_precision(
    value,
    round_floor=False,
    precision=2,
    dialect_precision=5
):
    """
    floor a value in its int representation:
        >>> floor_to_precision(296999)
        297000

        amounts are of the form : value * 10 ** dialect_precision it allows to
        store dialect_precision numbers after comma for intermediary amounts for
        totals we want precision numbers

    :param int value: The value to floor
    :param bool round_floor: Should be rounded down ?
    :param int precision: How much significant numbers we want ?
    :param int dialect_precision: The number of zeros that are concerning the
    floatting part of our value
    """
    if not isinstance(value, Decimal):
        value = Decimal(str(value))

    dividor = 10 ** (dialect_precision - precision)

    value = value / Decimal(dividor)
    return floor(value, round_floor) * dividor


def floor(value, round_floor=False):
    """
        floor a float value
        :param value: float value to be rounded
        :param bool round_floor: Should the data be floor rounded ?
        :return: an integer

        >>> floor(296.9999999)
        297
    """
    if not isinstance(value, Decimal):
        value = Decimal(str(value))
    return int(dec_round(value, 0, round_floor))


def dec_round(dec, precision, round_floor=False):
    """
    Return a decimal object rounded to precision

    :param int precision: the number of decimals we want after the comma
    :param bool round_floor: Should the data be floor rounded ?
    """
    if round_floor:
        method = ROUND_DOWN
    else:
        method = ROUND_HALF_UP
    # On construit un nombre qui a le même nombre de 0 après la virgule que
    # ce que l'on veut en définitive
    precision_reference_tmpl = "%%.%df" % precision
    precision_reference = precision_reference_tmpl % 1
    precision = Decimal(precision_reference)
    return dec.quantize(precision, method)


def round(float_, precision, round_floor=False):
    """
    Return a float object rounded to precision
    :param float float_: the object to round
    :param int precision: the number of decimals we want after the comma
    :param bool round_floor: Should the data be floor rounded ?
    """
    dec = Decimal(float_)
    return float(dec_round(dec, precision, round_floor))


def amount(value, precision=2):
    """
        Convert a float value as an integer amount to store it in a database
        :param value: float value to convert
        :param precision: number of dot translation to make

        >>> amount(195.65)
        19565
    """
    converter = math.pow(10, precision)
    result = floor(value * converter)
    return result


def integer_to_amount(value, precision=2):
    """
    Convert an integer value to a float with precision numbers after comma
    """
    flat_point = Decimal(str(math.pow(10, -precision)))
    val = Decimal(str(value)) * flat_point
    return float(Decimal(str(val)).quantize(flat_point, ROUND_HALF_UP))


def percentage(value, _percent):
    """
        Return the value of the "percent" percent of the original "value"
    """
    return int(float(value) * (float(_percent)/100.0))


def percent(part, total, default=None):
    """
        Return the percentage of total represented by part
        if default is provided, the ZeroDivisionError is handled
    """
    if default is not None and total == 0:
        return default
    value = part * 100.0 / total
    return float(dec_round(Decimal(str(value)), 2))


def convert_to_int(value, default=None):
    """
    try to convert the given value to an int
    """
    try:
        val = int(value)
    except ValueError as err:
        if default:
            val = default
        else:
            raise err
    return val


def convert_to_float(value, default=None):
    """
    Try to convert the given value object to a float
    """
    if isinstance(value, (str, bytes)):
        value = value.replace(',', '.')
    try:
        val = float(value)
    except ValueError as err:
        if default is not None:
            val = default
        else:
            raise err
    return val


# TVA related functions
def reverse_tva(total_ttc, tva, float_format=True):
    """
    Compute total_ht from total_ttc

    :param float total_ttc: ttc value in float format (by default)
    :param integer tva: the tva value in integer format (tva * 100)
    :param bool float_format: Is total_ttc in the float format (real ttc value)

    :returns: the value in integer format
    """
    # e.g : tva = 19.6 * 100 = 1960
    tva_dividor = max(int(tva), 0) + 100 * 100.0

    # First we translate the float value to an integer representation
    if float_format:
        total_ttc = amount(total_ttc, precision=5)

    # Representation in the integer representation
    result = floor(total_ttc * 10000 / tva_dividor)

    # We translate the result back to a float value
    if float_format:
        result = integer_to_amount(result, precision=5)

    return result


def compute_tva(total_ht, tva):
    """
        Compute the tva for the given ht total
    """
    return float(total_ht) * (max(int(tva), 0) / 10000.0)



"""
    Initialization functions
"""
import warnings
import logging

from sqlalchemy.exc import SAWarning

from endi_base.models.base import DBSESSION
from endi_base.models.base import DBBASE


def initialize_sql(engine):
    """
    Initialize the database engine
    """
    logger = logging.getLogger(__name__)
    logger.debug("Setting the metadatas")
    DBSESSION.configure(bind=engine)
    DBBASE.metadata.bind = engine
    return DBSESSION


def configure_warnings():
    """
    Python warning system setup

    Turns the sqla warning about implicit cartesian product into an exception,
    to be sure not to miss'em.

    If cartesian product is intentional, make it explicit.
    https://docs.sqlalchemy.org/en/14/changelog/migration_14.html#change-4737
    """
    warnings.filterwarnings(
        "error",
        category=SAWarning,
        # module='sqlalchemy.orm.relationships'
        module="sqlalchemy.sql.compiler",
        message=".*cartesian.*",
    )
